#include <unistd.h>
#include "mpi.h"
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <iostream>
#include <fstream>
#include "ti/fftlib/src/common/fft_common.h"

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>
#include "ocl_util.h"

#ifndef PI
#define PI 3.141592654
#endif

/* ======================================================================== */
/*  Parameters of fixed dataset.                                            */
/* ======================================================================== */
#define MAXN  (8*1024) /* Maximum size allowed per batch */
#define FFTSZ (MAXN)  /* Size of the FFT per batch */
#define BATCH (64) /* 8 batches of 8K each */
#define SPSIZE(qty) ((qty)*(sizeof(float)))
#define FFT_FORWARD 0

#define FFTBUF_X_WORK_SIZE (8*MAXN)
#define FFTBUF_Y_WORK_SIZE (16*MAXN)
#define FFTBUF_W_WORK_SIZE (6*MAXN + 256)
#define FFTBUF_BN_SIZE (4*MAXN)
#define FFTBUF_WORKBUF_SIZE (8*MAXN)
#define FFTBUF_POOL_SIZE  (FFTBUF_X_WORK_SIZE + FFTBUF_Y_WORK_SIZE + FFTBUF_W_WORK_SIZE + FFTBUF_BN_SIZE + FFTBUF_WORKBUF_SIZE)
#define FFTBUF_STATE_SIZE  0x1000

char fft_debug_outfile_name[256];
FILE *fft_debug_fp;

using namespace cl;
using namespace std;

#define FFTLIB_DEBUG_BUF_SIZE   0x1000
#define FFT_DEBUG_OUTFILE_BASE   "debug_out_%d_%02d.csv"

#define FFTLIB_NUM_PLANS   8

int fft_batch(Context& cxt, CommandQueue& cq, Program& prog, Buffer& buf_x, Buffer& buf_y, Buffer& buf_state, int N, int batch, char *processor_name)
{
     Buffer *bufs_w[FFTLIB_NUM_PLANS], *bufs_plan[FFTLIB_NUM_PLANS];
     int    single_size = 6*N, batch_per_plan = batch/FFTLIB_NUM_PLANS;
     unsigned int *state_ptr;
     cl_ulong t_strt, t_strt_min, t_end, t_end_max, t_execute = 0;

     /*------------------------------------------------------------------------
     * Create 3 Kernels from program, each with a kernel function name
     *-----------------------------------------------------------------------*/
     Kernel FFT_plan        (prog, "ocl_FFT_plan");
     Kernel FFT_execute     (prog, "ocl_FFT_execute");
     Kernel FFT_destroy_plan(prog, "ocl_FFT_destroy_plan");

     for (int i=0; i<FFTLIB_NUM_PLANS; i++)
     {
       /* Create DDR buffers, Kernels only read from them */
       bufs_w[i] = new Buffer(cxt, CL_MEM_READ_ONLY, (single_size + 256));
       /* Create MSMC buffers (limited in size, faster than DDR), Kernels read from and write to them */
       bufs_plan[i] = new Buffer(cxt, CL_MEM_READ_WRITE|CL_MEM_USE_MSMC_TI, sizeof(fft_plan_t));
     }

     /*------------------------------------------------------------------------
     * Set function arguments to FFT_plan kernle (ocl_FFT_plan() function)
     * Arguments set here are of indices: 0, 2, 3, 8, 9
     * Remaining arguments are set up later, before calling kernel function
     * __local(size) allocates scratch space in L2 for a __local argument
     * setArg(9, buf_state) will make function argument use memory in buf_state
     *-----------------------------------------------------------------------*/
     FFT_plan.setArg(0, N);
     FFT_plan.setArg(2, (int)FFT_FORWARD);
     FFT_plan.setArg(3, (int)FFT_ECPY);
     FFT_plan.setArg(8, __local(FFTBUF_POOL_SIZE));
     FFT_plan.setArg(9, buf_state);

     /*------------------------------------------------------------------------
     * Set function arguments to FFT_execute kernle (ocl_FFT_execute() function)
     * Arguments set here are of indices: 1, 2
     * Remaining arguments are set up later, before calling kernel function
     *-----------------------------------------------------------------------*/
     FFT_execute.setArg(1, __local(FFTBUF_POOL_SIZE));
     FFT_execute.setArg(2, buf_state);

     /*------------------------------------------------------------------------
     * Set function arguments to FFT_execute kernle (ocl_FFT_execute() function)
     * Arguments set here is of index: 1
     * Remaining arguments are set up later, before calling kernel function
     *-----------------------------------------------------------------------*/
     FFT_destroy_plan.setArg(1, buf_state);

     Event ev[FFTLIB_NUM_PLANS];

     /* Create FFT batch plans */
     for(int i=0; i<FFTLIB_NUM_PLANS; i++)
     {
       int              buflen   = N*batch_per_plan*sizeof(float);
       /* Defines (offset, size) for creating SubBuffer */
       cl_buffer_region data_x   = {i * buflen, buflen };
       cl_buffer_region data_y   = {2*i * buflen, 2*buflen };

       /* Arguments 1,4,7 of FFT_plan are set up here */
       FFT_plan.setArg(1, batch_per_plan);
       FFT_plan.setArg(4, *(bufs_plan[i]));
       FFT_plan.setArg(7, *(bufs_w[i]));

       /* Create SubBuffer of buf_x to be READ by Kernel.  No additional storage
        * is allocated, SubBuffers starts at offset of buf_x, for size long */
       Buffer x = buf_x.createSubBuffer(CL_MEM_READ_ONLY,  
                          CL_BUFFER_CREATE_TYPE_REGION, &data_x);
       /* Create SubBuffer of buf_y to be WRITTEN by Kernel. */
       Buffer y = buf_y.createSubBuffer(CL_MEM_WRITE_ONLY, 
                          CL_BUFFER_CREATE_TYPE_REGION, &data_y);

       /* Remaining arguments of FFT_plan are set up here */
       FFT_plan.setArg(5, x);
       FFT_plan.setArg(6, y);

       /* Dispatch FFT_plan kernel, track it with an event */
       cq.enqueueTask(FFT_plan,         0, &ev[i]);

       /* Wait till previous FFT_plan finishes */
       ev[i].wait();
     }

     /* Dispatch all execute tasks */
     for (int i=0; i<FFTLIB_NUM_PLANS; i++)
     {
       /* Remaining arguments of FFT_execute are set up here */
       FFT_execute.setArg(0, *bufs_plan[i]);
       /* Dispatch FFT_execute kernel, track it with an event */
       /* Note that cq was previously set to be OUT OF ORDER exeuction enabled,
          these FFT_execute kernels can run on multiple device cores
          simultaneously. */
       cq.enqueueTask(FFT_execute, 0, &ev[i]);
     }

     /* Wait for execute to complete, collect profiling stats, and destroy plans. */
     for (int i=0; i < FFTLIB_NUM_PLANS; i++)
     {
       ev[i].wait();

       ev[i].getProfilingInfo(CL_PROFILING_COMMAND_START,  &t_strt);
       ev[i].getProfilingInfo(CL_PROFILING_COMMAND_END,    &t_end);

       if ( i == 0 )
       {
         t_strt_min  = t_strt;
         t_end_max   = t_end;
       }
       else
       {
         if ( t_strt < t_strt_min ) t_strt_min = t_strt;
         if ( t_end  > t_end_max  ) t_end_max  = t_end;
       }
       t_execute += (t_end - t_strt);

       FFT_destroy_plan.setArg(0, *bufs_plan[i]);
       cq.enqueueTask(FFT_destroy_plan,0, &ev[i]);

       ev[i].wait();

       /* Deleting buffer objects, releasing buffer memory */
       delete bufs_w[i];
       delete bufs_plan[i];
     }

//     printf("total time = %8.6f, summed time = %8.6f\n", (t_end_max - t_strt_min)/1e6, t_execute/1e6);
     return (t_end_max - t_strt_min); //t_execute;
}



/* Output format string to store real and imaginary part of the output */
char csv_output[FFTSZ*BATCH][50];

/* ======================================================================== */
/*  MAIN -- Top level driver for the test.                                  */
/* ======================================================================== */
int main  (int argc, char *argv[])
{
  int i, j;
  int rad = 4;
  int rank, num_processes;
  char processor_name[320];
  int name_len;
  int errs = 0;
  int size, rc, *buf_in, *buf_out, count;
  MPI_File fh_in, fh_out;
  MPI_Comm comm;
  MPI_Status status;
  uint32_t nodeBits,flags;

  /* Read parameters from command line */
   if(argc < 3)
  {
    printf("Usage: %s <input_file> <output_file>\n",argv[0]);
    exit(-1);
  }
 /* Initiate the MPI process over multiple nodes. In this demo there is one MPI process per chip */
  MPI_Init( &argc, &argv );
  comm = MPI_COMM_WORLD;
  MPI_Comm_size( comm, &num_processes );
  MPI_Comm_rank( comm, &rank );
  int total_size,chunks_per_chip,num_chips,num_iterations;

  /* Get the name of the processor */
  MPI_Get_processor_name(processor_name, &name_len);
  gethostname(processor_name, 320);

  printf ("\n %s:.... FFT execution begins.....", processor_name);
  fflush(stdout);

  /* Test the input file */
  MPI_File_open( comm, argv[1], MPI_MODE_RDONLY, MPI_INFO_NULL, &fh_in );
  MPI_File_open( comm, argv[2], MPI_MODE_WRONLY | MPI_MODE_CREATE, MPI_INFO_NULL, &fh_out );


   try
   {
     Context context(CL_DEVICE_TYPE_ACCELERATOR);
     std::vector<Device> devices = context.getInfo<CL_CONTEXT_DEVICES>();
     /*------------------------------------------------------------------------
     * Create a command queue, note that OUT OF ORDER execution mode is enabled
     *-----------------------------------------------------------------------*/
     CommandQueue Q(context, devices[0], CL_QUEUE_PROFILING_ENABLE|CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE);

     /*------------------------------------------------------------------------
     * Read the offline compiled kernel module
     *-----------------------------------------------------------------------*/
     char *bin;
     int bin_length = ocl_read_binary("../c66x/kernel.out", bin);

     Program::Binaries   binary(1, std::make_pair(bin, bin_length));
     Program             program = Program(context, devices, binary);
     program.build(devices);

     delete [] bin;

     /*------------------------------------------------------------------------
     * Define the various buffers
     *-----------------------------------------------------------------------*/
     int dataSz    = SPSIZE(2*MAXN*BATCH);
     int twiddleSz = SPSIZE(2*MAXN);

     /* Kernel will only read from this DDR buffer */
     Buffer buf_x_b       (context, CL_MEM_READ_ONLY,  dataSz);
     /* Kernel will read from and write to this DDR buffer */
     Buffer buf_y_b       (context, CL_MEM_READ_WRITE, dataSz);
     /* Kernel will read from and write to this MSMC buffer (limited in size but faster than DDR) */
     Buffer buf_edma_state(context, CL_MEM_READ_WRITE|CL_MEM_USE_MSMC_TI,FFTBUF_STATE_SIZE);

     /* Map this MSMC buffer into host's virtual address space for WRITE. CL_TRUE means this call will return only after MapBuffer finishes */
     unsigned int *state_ptr = (unsigned int *)Q.enqueueMapBuffer(buf_edma_state, CL_TRUE, CL_MAP_WRITE, 0, FFTBUF_STATE_SIZE);
     memset(state_ptr,0,FFTBUF_STATE_SIZE);
     /* Unmap this MSMC buffer (write data back to device), free the virtual address space on host */
     Q.enqueueUnmapMemObject(buf_edma_state, state_ptr);

     bool continue_computing=true;
     int iteration=0;
     /* The input file is read in terms of 64K sized blocks of data, and dispatched to the DSP for
        FFT processing. If there are 'n' DSPs avilable, n*64K data is read from the input file and
       64K blocks are dispatched to the n DSPs for concurrent processing. The dispatching mechanism
       happens via MPI processes, one of which run on each of the n DSPs. Those MPI processes, with
       the help of OpenCL executes the FFT algorithm on its DSP cores and store the result in a shared
       file.
       The input file is shared across the MPI processes. As mentioned earlier, a batch of n*64K bytes
       are read from the input file. The file read pointer for an MPI process is moved to the beginning
       of the 64K data chunk, which it will read and perform FFT on. (The 'rank' gives the index of the
       process among the 'n' processes (num_processes) which could be used to advance the file pointer)
       Once  the execution is done, the output is written to the output file by the processes, in its
       respective positions within the file.
       This iteration is continued until the end of file is reached.

       The 64K batch given to DSP will be split in to 8 independent FFTS of 8Ks each internally by the 
       BATCH FFT process.
     */

    MPI_Barrier( comm );
     /*------------------------------------------------------------------------
     * For various sizes of FFT
     *-----------------------------------------------------------------------*/
     while(1)
     {
        /*---------------------------------------------------------------------
        * Map these two DDR buffers into host's virtual address space for WRITE,
        * blocking call until MapBuffer finishes
        *--------------------------------------------------------------------*/
        float *x_b = (float*)Q.enqueueMapBuffer(buf_x_b, CL_TRUE, CL_MAP_WRITE, 0, dataSz);
        float *y_b = (float*)Q.enqueueMapBuffer(buf_y_b, CL_TRUE, CL_MAP_WRITE, 0, dataSz);

     
        /* Set the file pointers to the position within the num_processes*64K block */
       MPI_File_seek(fh_in, (iteration* num_processes + rank) * sizeof(float) * FFTSZ*BATCH, MPI_SEEK_SET);

       memset( &status, 0xff, sizeof(MPI_Status) );
       MPI_File_read( fh_in, x_b, FFTSZ*BATCH, MPI_FLOAT, &status );
       MPI_Get_count( &status, MPI_INT, &count );

       if(count<(FFTSZ*BATCH)) {
         printf("\n %s: End of file reached, after %d iterations",processor_name,iteration);
         break;
       }

        /* ---------------------------------------------------------------- */
        /* Force uninitialized arrays to fixed values.                      */
        /* ---------------------------------------------------------------- */
        memset ((void*)y_b, 0xA5, dataSz);

        
        /* ------------------------------------------------------------------- 
        * Unmap/Release the buffers (write data back to device),
        * free virtual address spaces on host
        *--------------------------------------------------------------------*/
        Q.enqueueUnmapMemObject(buf_x_b, x_b);
        Q.enqueueUnmapMemObject(buf_y_b, y_b);

        Q.finish(); // wait for all unmaps to finish

        /*---------------------------------------------------------------------
        *--------------------------------------------------------------------*/

        cl_ulong t_execute;
        t_execute = fft_batch(context, Q, program, buf_x_b, buf_y_b, buf_edma_state, FFTSZ, BATCH, processor_name);

        /*---------------------------------------------------------------------
        * Map output buffers for READ, so results can be checked.
        *--------------------------------------------------------------------*/
        y_b = (float*)Q.enqueueMapBuffer(buf_y_b, CL_TRUE, CL_MAP_READ, 0, dataSz);

        cout << "\n"<< processor_name <<":" << " Batch FFTW executed: " << (t_execute/1000) << " us, iteration: "<<iteration<< endl;

#if 0
       /* The output (out[]) is a FFTSZ sized floating point pair or real & imaginary values (x,y).
          THis output (out[]) is formated in to strings containing the comma seperated values of
          the real and imaginary part and these output strings are stored in to the outputfile using MPI */
        for(i=0;i<(FFTSZ*BATCH);i++) {
          sprintf(csv_output[i],"%e,%e\n",y_b[2*i],y_b[2*i + 1]);
        }

       /* Seting the file pointer to the position within the output file */
       MPI_File_seek(fh_out, (iteration* num_processes + rank) * sizeof(csv_output), MPI_SEEK_SET );

       /* Store the comma seperated formatted fft output in to the output file */
       MPI_File_write( fh_out, csv_output,sizeof(csv_output), MPI_CHAR, &status );
#else
       /* Seting the file pointer to the position within the output file */
       MPI_File_seek(fh_out, (iteration* num_processes + rank) * sizeof(float) * (2*FFTSZ*BATCH), MPI_SEEK_SET );

       /* Store the comma seperated formatted fft output in to the output file */
       MPI_File_write( fh_out, y_b,(2*FFTSZ*BATCH), MPI_FLOAT, &status );
#endif
       iteration++;

        /* ------------------------------------------------------------------- 
        * Unmap/Release the buffers, free the virtual address space on host
        *--------------------------------------------------------------------*/
        Q.enqueueUnmapMemObject(buf_y_b, y_b);
        Q.finish(); // wait for all unmaps to finish


     }
        printf("\n %s .....Batch FFT execution completed....\n", processor_name);
   }
   catch (Error err)
   { cerr << "ERROR: " << err.what() << "(" << err.err() << ")" << endl; }

  /* Making sure all processes are completed up to this point */
    MPI_Barrier( comm );

    MPI_File_close( &fh_in );
    MPI_File_close( &fh_out );
    MPI_Finalize();
}














